"""Naive estimators."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import numpy as np
import numpy.typing as npt
import sklearn.utils.metadata_routing as skm
import sklearn.utils.validation as skv

import skfolio.typing as skt
from skfolio.optimization._base import BaseOptimization
from skfolio.prior import BasePrior, EmpiricalPrior
from skfolio.utils.stats import rand_weights_dirichlet
from skfolio.utils.tools import check_estimator


class InverseVolatility(BaseOptimization):
    """Inverse Volatility estimator.

    Each asset weight is computed using the inverse of its volatility and rescaled to
    have a sum of weights equal to one. The assets volatilities are derived from the
    prior estimator's covariance matrix.

    Parameters
    ----------
    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.ReturnDistribution`
        containing the estimation of assets expected returns, covariance matrix,
        returns and Cholesky decomposition of the covariance.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        and `previous_weights`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"`
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary
        estimator so that `fit` still returns the original instance. For traceability,
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets,), optional
        When `fallback="previous_weights"`, failures will fall back to these weights
        if provided.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.
    """

    prior_estimator_: BasePrior

    def __init__(
        self,
        prior_estimator: BasePrior | None = None,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        previous_weights: skt.MultiInput | None = None,
        raise_on_failure: bool = True,
    ):
        super().__init__(
            portfolio_params=portfolio_params,
            fallback=fallback,
            previous_weights=previous_weights,
            raise_on_failure=raise_on_failure,
        )
        self.prior_estimator = prior_estimator

    def get_metadata_routing(self):
        # noinspection PyTypeChecker
        router = skm.MetadataRouter(owner=self.__class__.__name__).add(
            prior_estimator=self.prior_estimator,
            method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
        )
        return router

    def fit(
        self, X: npt.ArrayLike, y: npt.ArrayLike | None = None, **fit_params
    ) -> InverseVolatility:
        """Fit the Inverse Volatility estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
           Price returns of the assets.

        y : array-like of shape (n_observations, n_targets), optional
            Price returns of factors or a target benchmark.
            The default is `None`.

        **fit_params : dict
            Parameters to pass to the underlying estimators.
            Only available if `enable_metadata_routing=True`, which can be
            set by using ``sklearn.set_config(enable_metadata_routing=True)``.
            See :ref:`Metadata Routing User Guide <metadata_routing>` for
            more details.

        Returns
        -------
        self : InverseVolatility
            Fitted estimator.
        """
        routed_params = skm.process_routing(self, "fit", **fit_params)

        # fitting prior estimator
        self.prior_estimator_ = check_estimator(
            self.prior_estimator,
            default=EmpiricalPrior(),
            check_type=BasePrior,
        )
        self.prior_estimator_.fit(X, y, **routed_params.prior_estimator.fit)
        covariance = self.prior_estimator_.return_distribution_.covariance
        w = 1 / np.sqrt(np.diag(covariance))
        self.weights_ = w / sum(w)
        return self


class EqualWeighted(BaseOptimization):
    """Equally Weighted estimator.

    Each asset weight is equal to `1/n_assets`.

    Parameters
    ----------
    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        and `previous_weights`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"`
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary
        estimator so that `fit` still returns the original instance. For traceability,
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets,), optional
        When `fallback="previous_weights"`, failures will fall back to these weights
        if provided.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.
    """

    def __init__(
        self,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        previous_weights: skt.MultiInput | None = None,
        raise_on_failure: bool = True,
    ):
        super().__init__(
            portfolio_params=portfolio_params,
            fallback=fallback,
            previous_weights=previous_weights,
            raise_on_failure=raise_on_failure,
        )

    def fit(self, X: npt.ArrayLike, y=None) -> EqualWeighted:
        """Fit the Equal Weighted estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : EqualWeighted
            Fitted estimator.
        """
        X = skv.validate_data(self, X)
        n_assets = X.shape[1]
        self.weights_ = np.ones(n_assets) / n_assets
        return self


class Random(BaseOptimization):
    """Random weight estimator.

    The assets weight are drawn from a Dirichlet distribution and sum to one.

    Parameters
    ----------
    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        and `previous_weights`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"`
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary
        estimator so that `fit` still returns the original instance. For traceability,
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets,), optional
        When `fallback="previous_weights"`, failures will fall back to these weights
        if provided.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,) or (n_optimizations, n_assets)
        Weights of the assets.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.
    """

    def __init__(
        self,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        previous_weights: skt.MultiInput | None = None,
        raise_on_failure: bool = True,
    ):
        super().__init__(
            portfolio_params=portfolio_params,
            fallback=fallback,
            previous_weights=previous_weights,
            raise_on_failure=raise_on_failure,
        )

    def fit(self, X: npt.ArrayLike, y=None):
        """Fit the Random Weighted estimator.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : EqualWeighted
            Fitted estimator.
        """
        X = skv.validate_data(self, X)
        n_assets = X.shape[1]
        self.weights_ = rand_weights_dirichlet(n=n_assets)
        return self



"""pre-selection estimators module."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Implementation derived from:
# Conway-Yu https://github.com/skfolio/skfolio/discussions/60
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt

import numpy as np
import pandas as pd
import sklearn.base as skb
import sklearn.feature_selection as skf
import sklearn.utils.validation as skv


class SelectNonExpiring(skf.SelectorMixin, skb.BaseEstimator):
    """
    Transformer to select assets that do not expire within a specified lookahead period
    after the end of the observation period.

    This transformer removes assets (columns) that have expiration dates within a
    given lookahead period from the end of the dataset, allowing only assets that
    remain active beyond this lookahead period to be selected.

    This is useful when an exit strategy is needed before asset expiration, such as
    for bonds or options with known end dates, or when applying WalkForward
    cross-validation. It ensures that assets expiring during the test period are
    excluded, so that only live assets are included in each training and test period.

    Parameters
    ----------
    expiration_dates : dict[str, dt.datetime | pd.Timestamp], optional
        Dictionary with asset names as keys and expiration dates as values.
        Used to check if each asset expires within the date offset.
        Assets with no expiration date will be retained by default.

    expiration_lookahead : pd.offsets.BaseOffset | dt.timedelta, optional
        The lookahead period after the end of the dataset within which assets with
        expiration dates will be removed.

    Attributes
    ----------
    to_keep_ : ndarray of shape (n_assets, )
       Boolean array indicating which assets are remaining.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    Notes
    -----
    This transformer only supports DataFrames with a DateTime index.

    Examples
    --------
    >>> import pandas as pd
    >>> import datetime as dt
    >>> from sklearn import set_config
    >>> set_config(transform_output="pandas")
    >>> X = pd.DataFrame(
    ...    {
    ...        'asset1': [1, 2, 3, 4],
    ...        'asset2': [2, 3, 4, 5],
    ...        'asset3': [3, 4, 5, 6],
    ...        'asset4': [4, 5, 6, 7]
    ...    }, index=pd.date_range("2023-01-01", periods=4, freq="D")
    ...)
    >>> expiration_dates = {
    ...    'asset1': pd.Timestamp("2023-01-10"),
    ...    'asset2': pd.Timestamp("2023-01-02"),
    ...    'asset3': pd.Timestamp("2023-01-06"),
    ...    'asset4': dt.datetime(2023, 5, 1)
    ... }
    >>> selector = SelectNonExpiring(
    ...    expiration_dates=expiration_dates,
    ...    expiration_lookahead=pd.DateOffset(days=5)
    ...)
    >>> selector.fit_transform(X)
               asset1  asset4
    2023-01-01      1      4
    2023-01-02      2      5
    2023-01-03      3      6
    2023-01-04      4      7
    """

    to_keep_: np.ndarray

    def __init__(
        self,
        expiration_dates: dict[str, dt.datetime | pd.Timestamp] | None = None,
        expiration_lookahead: pd.offsets.BaseOffset | dt.timedelta | None = None,
    ):
        self.expiration_dates = expiration_dates
        self.expiration_lookahead = expiration_lookahead

    def fit(self, X: pd.DataFrame, y=None) -> "SelectNonExpiring":
        """Run the SelectNonExpiring transformer and get the appropriate assets.

        Parameters
        ----------
        X : pd.DataFrame of shape (n_observations, n_assets)
            Returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : SelectNonExpiring
            Fitted estimator.
        """
        _ = skv.validate_data(self, X, ensure_all_finite="allow-nan")

        # Validate by allowing NaNs
        if not hasattr(X, "index") or not isinstance(X.index, pd.DatetimeIndex):
            raise ValueError(
                "X must be a DataFrame with an index of type DatetimeIndex"
            )

        if self.expiration_dates is None:
            raise ValueError("`expiration_dates` must be provided")

        if self.expiration_lookahead is None:
            raise ValueError("`expiration_lookahead` must be provided")

        # Calculate the cutoff date
        end_date = X.index[-1]
        cutoff_date = end_date + self.expiration_lookahead
        self.to_keep_ = np.array(
            [
                self.expiration_dates.get(asset, pd.Timestamp.max) > cutoff_date
                for asset in X.columns
            ]
        )

        return self

    def _get_support_mask(self) -> np.ndarray:
        skv.check_is_fitted(self)
        return self.to_keep_

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.input_tags.allow_nan = True
        return tags



"""Hierarchical Clustering estimators."""

# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

from skfolio.cluster._hierarchical import HierarchicalClustering, LinkageMethod

__all__ = ["HierarchicalClustering", "LinkageMethod"]



"""
The :mod:`skfolio.exceptions` module includes all custom warnings and error
classes used across skfolio.
"""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause

__all__ = [
    "DuplicateGroupsError",
    "EquationToMatrixError",
    "GroupNotFoundError",
    "NonPositiveVarianceError",
    "OptimizationError",
    "SolverError",
]


class OptimizationError(Exception):
    """Optimization Did not converge."""


class SolverError(Exception):
    """Solver error."""


class EquationToMatrixError(Exception):
    """Error while processing equations."""


class GroupNotFoundError(Exception):
    """Group name not found in the groups."""


class DuplicateGroupsError(Exception):
    """Group name appear in multiple group levels."""


class NonPositiveVarianceError(Exception):
    """Variance negative or null."""



"""Bivariate Copula Selection."""

# Copyright (c) 2025
# Authors: The skfolio developers
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import numpy.typing as npt
import scipy.stats as st
import sklearn as sk

from skfolio.distribution._base import SelectionCriterion
from skfolio.distribution.copula._base import BaseBivariateCopula
from skfolio.distribution.copula._clayton import ClaytonCopula
from skfolio.distribution.copula._gaussian import GaussianCopula
from skfolio.distribution.copula._gumbel import GumbelCopula
from skfolio.distribution.copula._independent import IndependentCopula
from skfolio.distribution.copula._joe import JoeCopula
from skfolio.distribution.copula._student_t import StudentTCopula


def select_bivariate_copula(
    X: npt.ArrayLike,
    copula_candidates: list[BaseBivariateCopula] | None = None,
    selection_criterion: SelectionCriterion = SelectionCriterion.AIC,
    independence_level: float = 0.05,
) -> BaseBivariateCopula:
    """
    Select the best bivariate copula from a list of candidates using an information
    criterion.

    This function first tests the dependence between the two variables in X using
    Kendall's tau independence test. If the p-value is greater than or equal to
    `independence_level`, the null hypothesis of independence is not rejected, and the
    `IndependentCopula` is returned. Otherwise, each candidate copula in
    `copula_candidates` is fitted to the data X. For each candidate, either the
    Akaike Information Criterion (AIC) or the Bayesian Information Criterion (BIC) is
    computed, and the copula with the lowest criterion value is selected.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        An array of bivariate inputs (u, v) with uniform marginals (values in [0, 1]).

    copula_candidates : list[BaseBivariateCopula]
        A list of candidate copula models. Each candidate must inherit from
        `BaseBivariateCopula`. If None, defaults to
        `[GaussianCopula(), StudentTCopula(), ClaytonCopula(), GumbelCopula(), JoeCopula()]`.

    selection_criterion : SelectionCriterion, default=SelectionCriterion.AIC
        The criterion used for model selection. Possible values are:
            - SelectionCriterion.AIC : Akaike Information Criterion
            - SelectionCriterion.BIC : Bayesian Information Criterion

    independence_level : float, default=0.05
        The significance level for the Kendall tau independence test. If the p-value is
        greater than or equal to this level, the independence hypothesis is not
        rejected, and the `IndependentCopula` is returned.

    Returns
    -------
    selected_copula : BaseBivariateCopula
        The fitted copula model among the candidates that minimizes the selected
        information criterion (AIC or BIC).

    Raises
    ------
    ValueError
        If X is not a 2D array with exactly two columns, or if any candidate in
        `copula_candidates` does not inherit from `BaseBivariateCopula`.
    """
    if copula_candidates is None:
        copula_candidates = [
            GaussianCopula(),
            StudentTCopula(),
            ClaytonCopula(),
            GumbelCopula(),
            JoeCopula(),
        ]

    X = np.asarray(X)
    if X.ndim != 2 or X.shape[1] != 2:
        raise ValueError("X must contains two columns for Bivariate Copula")

    kendall_tau, p_value = st.kendalltau(X[:, 0], X[:, 1])
    if p_value >= independence_level:
        return IndependentCopula().fit(X)

    results = {}
    for copula in copula_candidates:
        if not isinstance(copula, BaseBivariateCopula):
            raise ValueError(
                "The candidate copula must inherit from `BaseBivariateCopula`"
            )
        copula = sk.clone(copula)
        if copula.itau and copula.kendall_tau is None:
            # Faster computation by reusing kendall tau if itau
            copula.kendall_tau = kendall_tau
        copula.fit(X)

        match selection_criterion:
            case selection_criterion.AIC:
                results[copula] = copula.aic(X)
            case selection_criterion.BIC:
                results[copula] = copula.bic(X)
            case _:
                raise ValueError(f"{selection_criterion} not implemented")

    selected_copula = min(results, key=results.get)
    return selected_copula



"""Copula module."""

from skfolio.distribution.copula._base import (
    UNIFORM_MARGINAL_EPSILON,
    BaseBivariateCopula,
)
from skfolio.distribution.copula._clayton import ClaytonCopula
from skfolio.distribution.copula._gaussian import GaussianCopula
from skfolio.distribution.copula._gumbel import GumbelCopula
from skfolio.distribution.copula._independent import IndependentCopula
from skfolio.distribution.copula._joe import JoeCopula
from skfolio.distribution.copula._selection import select_bivariate_copula
from skfolio.distribution.copula._student_t import StudentTCopula
from skfolio.distribution.copula._utils import (
    CopulaRotation,
    compute_pseudo_observations,
    empirical_tail_concentration,
    plot_tail_concentration,
)

__all__ = [
    "UNIFORM_MARGINAL_EPSILON",
    "BaseBivariateCopula",
    "ClaytonCopula",
    "CopulaRotation",
    "GaussianCopula",
    "GumbelCopula",
    "IndependentCopula",
    "JoeCopula",
    "StudentTCopula",
    "compute_pseudo_observations",
    "empirical_tail_concentration",
    "plot_tail_concentration",
    "select_bivariate_copula",
]



"""Multivariate Distribution module."""

from skfolio.distribution.multivariate._base import BaseMultivariateDist
from skfolio.distribution.multivariate._utils import DependenceMethod
from skfolio.distribution.multivariate._vine_copula import VineCopula

__all__ = [
    "BaseMultivariateDist",
    "DependenceMethod",
    "VineCopula",
]



"""Base Univariate Estimator."""

# Copyright (c) 2025
# Authors: The skfolio developers
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import warnings
from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import scipy.stats as st
import sklearn.utils as sku
import sklearn.utils.validation as skv

from skfolio.distribution._base import BaseDistribution


class BaseUnivariateDist(BaseDistribution, ABC):
    """Base Univariate Distribution Estimator.

    This abstract class serves as a foundation for univariate distribution models
    based on scipy.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.
    """

    _scipy_model: st.rv_continuous

    def __init__(self, random_state: int | None = None):
        super().__init__(random_state=random_state)

    @property
    @abstractmethod
    def _scipy_params(self) -> dict[str, float]:
        """Dictionary of parameters to pass to the underlying SciPy distribution."""
        pass

    @property
    def n_params(self) -> int:
        """Number of model parameters."""
        return len(self._scipy_params)

    @property
    def fitted_repr(self) -> str:
        """String representation of the fitted univariate distribution."""
        skv.check_is_fitted(self)
        params = ", ".join([f"{k}={v:0.2g}" for k, v in self._scipy_params.items()])
        return f"{self.__class__.__name__}({params})"

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y=None) -> "BaseUnivariateDist":
        """Fit the univariate distribution model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            The input data. X must contain a single column.


        y : None
            Ignored. Provided for compatibility with scikit-learn's API.

        Returns
        -------
        self : BaseUnivariateDist
            Returns the instance itself.
        """
        pass

    def _validate_X(self, X: npt.ArrayLike, reset: bool) -> np.ndarray:
        """Validate and convert the input data X.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            The input data. X must contain a single column.

        reset : bool, default=True
            Whether to reset the `n_features_in_` attribute.
            If False, the input will be checked for consistency with data
            provided when reset was last True.

        Returns
        -------
        validated_X : ndarray of shape (n_observations, 1).
            The validated input array
        """
        X = skv.validate_data(self, X, dtype=np.float64, reset=reset)
        if X.shape[1] != 1:
            raise ValueError(
                "X should contain a single column for Univariate Distribution"
            )
        return X

    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the log-likelihood of each sample (log-pdf) under the model.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            An array of points at which to evaluate the log-probability density.
            The data should be a single feature column.

        Returns
        -------
        density : ndarray of shape (n_observations,)
            Log-likelihood values for each observation in X.
        """
        X = self._validate_X(X, reset=False)
        log_density = self._scipy_model.logpdf(X, **self._scipy_params).ravel()
        return log_density

    def sample(self, n_samples: int = 1):
        """Generate random samples from the fitted distribution.

        Currently, this is implemented only for gaussian and tophat kernels.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to generate.

        Returns
        -------
        X : array-like of shape (n_samples, 1)
            List of samples.
        """
        skv.check_is_fitted(self)
        rng = sku.check_random_state(self.random_state)
        sample = self._scipy_model.rvs(
            size=(n_samples, 1), random_state=rng, **self._scipy_params
        )
        return sample

    def cdf(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the cumulative distribution function (CDF) for the given data.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            Data points at which to evaluate the CDF.

        Returns
        -------
        cdf : ndarray of shape (n_observations, 1)
            The CDF evaluated at each data point.
        """
        skv.check_is_fitted(self)
        return self._scipy_model.cdf(X, **self._scipy_params)

    def ppf(self, X: npt.ArrayLike) -> np.ndarray:
        """Compute the percent point function (inverse of the CDF) for the given
         probabilities.

        Parameters
        ----------
        X : array-like of shape (n_observations, 1)
            Probabilities for which to compute the corresponding quantiles.

        Returns
        -------
         ppf : ndarray of shape (n_observations, 1)
            The quantiles corresponding to the given probabilities.
        """
        skv.check_is_fitted(self)
        return self._scipy_model.ppf(X, **self._scipy_params)

    def plot_pdf(
        self, X: npt.ArrayLike | None = None, title: str | None = None
    ) -> go.Figure:
        """Plot the probability density function (PDF).

        Parameters
        ----------
        X : array-like of shape (n_samples, 1), optional
            If provided, it is used to plot the empirical data KDE for comparison
            versus the model PDF.

        title : str, optional
           The title for the plot. If not provided, a default title based on the fitted
           model's representation is used.

        Returns
        -------
        fig : go.Figure
           A Plotly figure object containing the PDF plot.
        """
        skv.check_is_fitted(self)
        if title is None:
            title = f"PDF of {self.__class__.__name__}"
            if X is not None:
                title += " vs Empirical KDE"

        # Compute the quantile-based range
        lower_bound = self.ppf(1e-4)
        upper_bound = self.ppf(1 - 1e-4)
        # Generate x values across this range
        x = np.linspace(lower_bound, upper_bound, 1000)

        traces = []

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            pdfs = np.exp(self.score_samples(x.reshape(-1, 1)))
        traces.append(
            go.Scatter(
                x=x,
                y=pdfs.flatten(),
                mode="lines",
                name=self.__class__.__name__,
                line=dict(color="rgb(31, 119, 180)", dash="solid", width=1),
                fill="tozeroy",
                fillcolor="rgba(31, 119, 180, 0.17)",
                opacity=1.0,
            )
        )

        if X is not None:
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message="^X has feature names", category=UserWarning
                )
                X = self._validate_X(X, reset=False)
            kde = st.gaussian_kde(X[:, 0])
            y_kde = kde(x)
            traces.append(
                go.Scatter(
                    x=x,
                    y=y_kde,
                    mode="lines",
                    name="Empirical KDE",
                    line=dict(color="rgb(85, 168, 104)", dash="dash", width=2),
                    fill="tozeroy",
                    fillcolor="rgba(85, 168, 104, 0.17)",
                    opacity=1.0,
                )
            )

        fig = go.Figure(data=traces)
        fig.update_layout(
            title=title,
            xaxis_title="x",
            yaxis_title="Probability Density",
        )
        fig.update_xaxes(
            tickformat=".0%",
        )
        return fig

    def qq_plot(self, X: npt.ArrayLike, title: str | None = None) -> go.Figure:
        """Plot the empirical quantiles of the sample X versus the quantiles of the
        fitted model.

        Parameters
        ----------
        X : array-like of shape (n_samples, 1), optional
            Used to plot the empirical quantiles for comparison versus the model
            quantiles.

        title : str, optional
           The title for the plot. If not provided, a default title based on the fitted
           model's representation is used.

        Returns
        -------
        fig : go.Figure
           A Plotly figure object containing the PDF plot.
        """
        skv.check_is_fitted(self)
        if title is None:
            title = f"Q-Q Plot of {self.__class__.__name__} vs Sample Data"

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="^X has feature names", category=UserWarning
            )
            X = self._validate_X(X, reset=False)

        X_sorted = np.sort(X[:, 0])
        n = len(X)

        # Compute theoretical quantiles from the model
        theoretical_quantiles = self.ppf((np.arange(1, n + 1) - 0.5) / n)

        # Create the Q-Q plot using Plotly
        fig = go.Figure(
            go.Scatter(
                x=theoretical_quantiles,
                y=X_sorted,
                mode="markers",
            )
        )
        # Add a reference line (45° line)
        min_val = min(float(theoretical_quantiles[0]), float(X_sorted[0]))
        max_val = max(float(theoretical_quantiles[-1]), float(X_sorted[-1]))
        fig.add_trace(
            go.Scatter(
                x=[min_val, max_val],
                y=[min_val, max_val],
                mode="lines",
            )
        )
        fig.update_layout(
            title=title,
            xaxis_title="Theoretical Quantiles",
            yaxis_title="Sample Quantiles",
            showlegend=False,
        )
        return fig



"""Prior module."""

from skfolio.prior._base import BasePrior, ReturnDistribution
from skfolio.prior._black_litterman import BlackLitterman
from skfolio.prior._empirical import EmpiricalPrior
from skfolio.prior._entropy_pooling import EntropyPooling
from skfolio.prior._factor_model import (
    BaseLoadingMatrix,
    FactorModel,
    LoadingMatrixRegression,
)
from skfolio.prior._opinion_pooling import OpinionPooling
from skfolio.prior._synthetic_data import SyntheticData

__all__ = [
    "BaseLoadingMatrix",
    "BasePrior",
    "BlackLitterman",
    "EmpiricalPrior",
    "EntropyPooling",
    "FactorModel",
    "LoadingMatrixRegression",
    "OpinionPooling",
    "ReturnDistribution",
    "SyntheticData",
]



"""Base Hierarchical Clustering Optimization estimator."""

# Copyright (c) 2023
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# SPDX-License-Identifier: BSD-3-Clause
# Implementation derived from:
# Riskfolio-Lib, Copyright (c) 2020-2023, Dany Cajas, Licensed under BSD 3 clause.
# scikit-learn, Copyright (c) 2007-2010 David Cournapeau, Fabian Pedregosa, Olivier

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import numpy.typing as npt
import sklearn.utils.metadata_routing as skm

import skfolio.typing as skt
from skfolio.cluster import HierarchicalClustering
from skfolio.distance import BaseDistance
from skfolio.measures import ExtraRiskMeasure, RiskMeasure
from skfolio.optimization._base import BaseOptimization
from skfolio.portfolio import Portfolio
from skfolio.prior import BasePrior, ReturnDistribution
from skfolio.utils.tools import input_to_array


class BaseHierarchicalOptimization(BaseOptimization, ABC):
    r"""Base Hierarchical Clustering Optimization estimator.

    Parameters
    ----------
    risk_measure : RiskMeasure or ExtraRiskMeasure, default=RiskMeasure.VARIANCE
        :class:`~skfolio.meta.RiskMeasure` or :class:`~skfolio.meta.ExtraRiskMeasure`
        of the optimization.
        Can be any of:

            * MEAN_ABSOLUTE_DEVIATION
            * FIRST_LOWER_PARTIAL_MOMENT
            * VARIANCE
            * SEMI_VARIANCE
            * CVAR
            * EVAR
            * WORST_REALIZATION
            * CDAR
            * MAX_DRAWDOWN
            * AVERAGE_DRAWDOWN
            * EDAR
            * ULCER_INDEX
            * GINI_MEAN_DIFFERENCE_RATIO
            * VALUE_AT_RISK
            * DRAWDOWN_AT_RISK
            * ENTROPIC_RISK_MEASURE
            * FOURTH_CENTRAL_MOMENT
            * FOURTH_LOWER_PARTIAL_MOMENT

        The default is `RiskMeasure.VARIANCE`.

    prior_estimator : BasePrior, optional
        :ref:`Prior estimator <prior>`.
        The prior estimator is used to estimate the :class:`~skfolio.prior.ReturnDistribution`
        containing the estimation of assets expected returns, covariance matrix and
        returns. The moments and returns estimations are used for the risk computation
        and the returns estimation are used by the distance matrix estimator.
        The default (`None`) is to use :class:`~skfolio.prior.EmpiricalPrior`.

    distance_estimator : BaseDistance, optional
        :ref:`Distance estimator <distance>`.
        The distance estimator is used to estimate the codependence and the distance
        matrix needed for the computation of the linkage matrix.
        The default (`None`) is to use :class:`~skfolio.distance.PearsonDistance`.

    hierarchical_clustering_estimator : HierarchicalClustering, optional
        :ref:`Hierarchical Clustering estimator <hierarchical_clustering>`.
        The hierarchical clustering estimator is used to compute the linkage matrix
        and the hierarchical clustering of the assets based on the distance matrix.
        The default (`None`) is to use
        :class:`~skfolio.cluster.HierarchicalClustering`.

    min_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Minimum assets weights (weights lower bounds). The default is 0.0 (no short
        selling). Negative weights are not allowed. If a float is provided, it is
        applied to each asset. `None` is equivalent to the default `0.0`. If a
        dictionary is provided, its (key/value) pair must be the (asset name/asset
        minium weight) and the input `X` of the `fit` methods must be a DataFrame with
        the asset names in columns. When using a dictionary, assets values that are not
        provided are assigned the default  minimum weight of `0.0`.

        Example:

           * `min_weights = 0.0` --> long only portfolio (default).
           * `min_weights = {"SX5E": 0.1, "SPX": 0.2}`
           * `min_weights = [0.1, 0.2]`

    max_weights : float | dict[str, float] | array-like of shape (n_assets, ), default=1.0
        Maximum assets weights (weights upper bounds). The default is 1.0 (each asset
        is below 100%). Weights above 1.0 are not allowed. If a float is provided, it is
        applied to each asset. `None` is equivalent to the default `1.0`. If a
        dictionary is provided, its (key/value) pair must be the (asset name/asset
        maximum weight) and the input `X` of the `fit` method must be a DataFrame with
        the asset names in columns. When using a dictionary, assets values that are not
        provided are assigned the default maximum weight of `1.0`.

        Example:

           * `max_weights = 1.0` --> each weight  must be below 100% (default).
           * `max_weights = 0.5` --> each weight must be below 50%.
           * `max_weights = {"SX5E": 0.8, "SPX": 0.9}`
           * `max_weights = [0.8, 0.9]`

    transaction_costs : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Transaction costs of the assets. It is used to add linear transaction costs to
        the optimization problem:

        .. math:: total\_cost = \sum_{i=1}^{N} c_{i} \times |w_{i} - w\_prev_{i}|

        with :math:`c_{i}` the transaction cost of asset i, :math:`w_{i}` its weight
        and :math:`w\_prev_{i}` its previous weight (defined in `previous_weights`).
        The float :math:`total\_cost` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_cost

        with :math:`\mu` the vector af assets' expected returns and :math:`w` the
        vector of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset cost) and the input `X` of the `fit` method must be a
        DataFrame with the asset names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the transaction costs
            needs to be homogenous to the periodicity of :math:`\mu`. For example, if
            the input `X` is composed of **daily** returns, the `transaction_costs` need
            to be expressed as **daily** costs.
            (See :ref:`sphx_glr_auto_examples_mean_risk_plot_6_transaction_costs.py`)

    management_fees : float | dict[str, float] | array-like of shape (n_assets, ), default=0.0
        Management fees of the assets. It is used to add linear management fees to the
        optimization problem:

        .. math:: total\_fee = \sum_{i=1}^{N} f_{i} \times w_{i}

        with :math:`f_{i}` the management fee of asset i and :math:`w_{i}` its weight.
        The float :math:`total\_fee` is impacting the portfolio expected return in the optimization:

        .. math:: expected\_return = \mu^{T} \cdot w - total\_fee

        with :math:`\mu` the vector af assets expected returns and :math:`w` the vector
        of assets weights.

        If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset fee) and the input `X` of the `fit` method must be a
        DataFrame with the asset names in columns.
        The default value is `0.0`.

        .. warning::

            Based on the above formula, the periodicity of the management fees needs to
            be homogenous to the periodicity of :math:`\mu`. For example, if the input
            `X` is composed of **daily** returns, the `management_fees` need to be
            expressed in **daily** fees.

        .. note::

            Another approach is to directly impact the management fees to the input `X`
            in order to express the returns net of fees. However, when estimating the
            :math:`\mu` parameter using for example Shrinkage estimators, this approach
            would mix a deterministic value with an uncertain one leading to unwanted
            bias in the management fees.

    previous_weights : float | dict[str, float] | array-like of shape (n_assets, ), optional
        Previous weights of the assets. Previous weights are used to compute the
        portfolio total cost. If a float is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/asset previous weight) and the input `X` of the `fit` method must
        be a DataFrame with the asset names in columns.
        The default (`None`) means no previous weights.
        Additionally, when `fallback="previous_weights"`, failures will fall back to
        these weights if provided.

    portfolio_params : dict, optional
        Portfolio parameters forwarded to the resulting `Portfolio` in `predict`.
        If not provided and if available on the estimator, the following
        attributes are propagated to the portfolio by default: `name`,
        `transaction_costs`, `management_fees`, `previous_weights` and `risk_free_rate`.

    fallback : BaseOptimization | "previous_weights" | list[BaseOptimization | "previous_weights"], optional
        Fallback estimator or a list of estimators to try, in order, when the primary
        optimization raises during `fit`. Alternatively, use `"previous_weights"`
        (alone or in a list) to fall back to the estimator's `previous_weights`.
        When a fallback succeeds, its fitted `weights_` are copied back to the primary
        estimator so that `fit` still returns the original instance. For traceability,
        `fallback_` stores the successful estimator (or the string `"previous_weights"`)
         and `fallback_chain_` stores each attempt with the associated outcome.

    raise_on_failure : bool, default=True
        Controls error handling when fitting fails.
        If True, any failure during `fit` is raised immediately, no `weights_` are
        set and subsequent calls to `predict` will raise a `NotFittedError`.
        If False, errors are not raised; instead, a warning is emitted, `weights_`
        is set to `None` and subsequent calls to `predict` will return a
        `FailedPortfolio`. When fallbacks are specified, this behavior applies only
        after all fallbacks have been exhausted.

    Attributes
    ----------
    weights_ : ndarray of shape (n_assets,)
        Weights of the assets.

    prior_estimator_ : BasePrior
        Fitted `prior_estimator`.

    distance_estimator_ : BaseDistance
        Fitted `distance_estimator`.

    hierarchical_clustering_estimator_ : HierarchicalClustering
        Fitted `hierarchical_clustering_estimator`.

    n_features_in_ : int
       Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
       Names of features seen during `fit`. Defined only when `X`
       has feature names that are all strings.

    fallback_ : BaseOptimization | "previous_weights" | None
        The fallback estimator instance, or the string `"previous_weights"`, that
        produced the final result. `None` if no fallback was used.

    fallback_chain_ : list[tuple[str, str]] | None
        Sequence describing the optimization fallback attempts. Each element is a
        pair `(estimator_repr, outcome)` where `estimator_repr` is the string
        representation of the primary estimator or a fallback (e.g. `"EqualWeighted()"`,
        `"previous_weights"`), and `outcome` is `"success"` if that step produced
        a valid solution, otherwise the stringified error message. For successful
        fits without any fallback, this is `None`.

    error_ : str | list[str] | None
        Captured error message(s) when `fit` fails. For multi-portfolio outputs
        (`weights_` is 2D), this is a list aligned with portfolios.

    Notes
    -----
    All estimators should specify all parameters as explicit keyword arguments in
    `__init__` (no `*args` or `**kwargs`), following scikit-learn conventions.
    """

    prior_estimator_: BasePrior
    distance_estimator_: BaseDistance
    hierarchical_clustering_estimator_: HierarchicalClustering

    @abstractmethod
    def __init__(
        self,
        risk_measure: RiskMeasure | ExtraRiskMeasure = RiskMeasure.VARIANCE,
        prior_estimator: BasePrior | None = None,
        distance_estimator: BaseDistance | None = None,
        hierarchical_clustering_estimator: HierarchicalClustering | None = None,
        min_weights: skt.MultiInput | None = 0.0,
        max_weights: skt.MultiInput | None = 1.0,
        transaction_costs: skt.MultiInput = 0.0,
        management_fees: skt.MultiInput = 0.0,
        previous_weights: skt.MultiInput | None = None,
        portfolio_params: dict | None = None,
        fallback: skt.Fallback = None,
        raise_on_failure: bool = True,
    ):
        super().__init__(
            portfolio_params=portfolio_params,
            fallback=fallback,
            previous_weights=previous_weights,
            raise_on_failure=raise_on_failure,
        )
        self.risk_measure = risk_measure
        self.prior_estimator = prior_estimator
        self.distance_estimator = distance_estimator
        self.hierarchical_clustering_estimator = hierarchical_clustering_estimator
        self.min_weights = min_weights
        self.max_weights = max_weights
        self.transaction_costs = transaction_costs
        self.management_fees = management_fees
        self._seriated = False

    def _clean_input(
        self,
        value: float | dict | np.ndarray | list,
        n_assets: int,
        fill_value: Any,
        name: str,
    ) -> np.ndarray:
        """Convert input to cleaned 1D array
         value : float, dict, array-like or None.
            Input value to clean and convert.

        Parameters
        ----------
        value : float, dict or array-like.
            Input value to clean.

        n_assets : int
            Number of assets. Used to verify the shape of the converted array.

        fill_value : Any
            When `items` is a dictionary, elements that are not in `asset_names` are
            filled with `fill_value` in the converted array.

        name : str
            Name used for error messages.

        Returns
        -------
        value :  ndarray of shape (n_assets,)
            The cleaned float or 1D array.
        """
        if value is None:
            raise ValueError("Cannot convert None to array")
        if np.isscalar(value):
            return value * np.ones(n_assets)
        return input_to_array(
            items=value,
            n_assets=n_assets,
            fill_value=fill_value,
            dim=1,
            assets_names=(
                self.feature_names_in_ if hasattr(self, "feature_names_in_") else None
            ),
            name=name,
        )

    def _risk(
        self,
        weights: np.ndarray,
        return_distribution: ReturnDistribution,
    ) -> float:
        """Compute the risk measure of a theoretical portfolio defined by the weights
        vector.

        Parameters
        ----------
        weights : ndarray of shape (n_assets,)
           The vector of weights.

        return_distribution : ReturnDistribution
            The assets return distribution.

        Returns
        -------
        risk: float
            The risk measure of a theoretical portfolio defined by the weights
            vector.
        """
        ptf = Portfolio(
            X=return_distribution.returns,
            sample_weight=return_distribution.sample_weight,
            weights=weights,
            transaction_costs=self.transaction_costs,
            management_fees=self.management_fees,
            previous_weights=self.previous_weights,
        )
        if self.risk_measure in [RiskMeasure.VARIANCE, RiskMeasure.STANDARD_DEVIATION]:
            risk = ptf.variance_from_assets(
                assets_covariance=return_distribution.covariance
            )
            if self.risk_measure == RiskMeasure.STANDARD_DEVIATION:
                risk = np.sqrt(risk)
        else:
            risk = getattr(ptf, str(self.risk_measure.value))
        return risk

    def _unitary_risks(self, return_distribution: ReturnDistribution) -> np.ndarray:
        """Compute the vector of risk measure for each single assets.

        Parameters
        ----------
        return_distribution : ReturnDistribution
            The asset returns distribution.

        Returns
        -------
        values: ndarray of shape (n_assets,)
            The risk measure of each asset.
        """
        n_assets = return_distribution.returns.shape[1]
        risks = [
            self._risk(weights=weights, return_distribution=return_distribution)
            for weights in np.identity(n_assets)
        ]
        return np.array(risks)

    def _convert_weights_bounds(self, n_assets: int) -> tuple[np.ndarray, np.ndarray]:
        """Convert the input weights lower and upper bounds to two 1D arrays.

        Parameters
        ----------
        n_assets : int
            Number of assets.

        Returns
        -------
        min_weights : ndarray of shape (n_assets,)
            The weight lower bound 1D array.
        max_weights : ndarray of shape (n_assets,)
            The weight upper bound 1D array.
        """
        if self.min_weights is None:
            min_weights = np.zeros(n_assets)
        else:
            min_weights = self._clean_input(
                self.min_weights,
                n_assets=n_assets,
                fill_value=0,
                name="min_weights",
            )
            if np.any(min_weights < 0):
                raise ValueError("`min_weights` must be strictly positive")
            if min_weights.sum() >= 1.00001:
                raise ValueError(
                    f"Invalid `min_weights`: sum is {min_weights.sum():.4f}, "
                    f"but it must be less than 1.0."
                )

        if self.max_weights is None:
            max_weights = np.ones(n_assets)
        else:
            max_weights = self._clean_input(
                self.max_weights,
                n_assets=n_assets,
                fill_value=1,
                name="max_weights",
            )
            if np.any(max_weights > 1):
                raise ValueError("`max_weights` must be less than or equal to 1.0")
            if max_weights.sum() < 1:
                raise ValueError(
                    f"Invalid `max_weights`: sum is {max_weights.sum():.4f}, "
                    f"but it must be at least 1.0."
                )

        if np.any(min_weights > max_weights):
            raise NameError(
                "Items of `min_weights` must be less than or equal to items of"
                " `max_weights`"
            )

        return min_weights, max_weights

    def get_metadata_routing(self):
        # noinspection PyTypeChecker
        router = (
            skm.MetadataRouter(owner=self.__class__.__name__)
            .add(
                prior_estimator=self.prior_estimator,
                method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
            )
            .add(
                distance_estimator=self.distance_estimator,
                method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
            )
            .add(
                hierarchical_clustering_estimator=self.hierarchical_clustering_estimator,
                method_mapping=skm.MethodMapping().add(caller="fit", callee="fit"),
            )
        )
        return router

    @abstractmethod
    def fit(self, X: npt.ArrayLike, y: None = None, **fit_params):
        pass



"""Bivariate Copula Utils."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

import operator
import warnings
from collections.abc import Callable
from enum import Enum

import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go
import scipy.optimize as so
import scipy.stats as st


class CopulaRotation(Enum):
    r"""Enum representing the rotation (in degrees) to apply to a bivariate copula.

    It follows the standard clockwise convention:

    - `CopulaRotation.R0` (0°): :math:`(u, v) \mapsto (u, v)`
    - `CopulaRotation.R90` (90°): :math:`(u, v) \mapsto (v,\, 1 - u)`
    - `CopulaRotation.R180` (180°): :math:`(u, v) \mapsto (1 - u,\, 1 - v)`
    - `CopulaRotation.R270` (270°): :math:`(u, v) \mapsto (1 - v,\, u)`

    Attributes
    ----------
    R0 : int
        No rotation (0°).
    R90 : int
        90° rotation.
    R180 : int
        180° rotation.
    R270 : int
        270° rotation.
    """

    R0 = "0°"
    R90 = "90°"
    R180 = "180°"
    R270 = "270°"

    def __str__(self) -> str:
        """String representation."""
        return self.value


def compute_pseudo_observations(X: npt.ArrayLike) -> np.ndarray:
    """
    Compute pseudo-observations by ranking each column of the data and scaling the
    ranks.

    The goal of computing pseudo-observations is to transform your raw data into a
    form that has uniform marginal distributions on the open interval (0, 1). This is
    particularly useful in copula modeling and other statistical methods where the
    dependence structure is of primary interest, independent of the marginal
    distributions.

    This function transforms each column of the input data into pseudo-observations
    on the (0, 1) interval. For each column, the ranks (starting at 1) are divided by
    (n_samples + 1) to avoid 0 and 1 values, which are problematic for many copula
    methods.

    Parameters
    ----------
    X : array-like of shape (n_observations, n_assets)
        Input data.

    Returns
    -------
    pseudo_observations: ndarray of shape (n_observations, n_assets)
        An array of pseudo-observations corresponding to the ranks scaled to (0, 1).
    """
    X = np.asarray(X)
    # Compute column-wise ranks; rankdata returns ranks starting at 1.
    ranks = st.rankdata(X, axis=0)
    n_samples = X.shape[0]
    pseudo_observations = ranks / (n_samples + 1)
    return pseudo_observations


def empirical_tail_concentration(X: npt.ArrayLike, quantiles: np.ndarray) -> np.ndarray:
    """
    Compute empirical tail concentration for the two variables in X.
    This function computes the concentration at each quantile provided.

    The tail concentration are estimated as:
      - Lower tail: λ_L(q) = P(U₂ ≤ q | U₁ ≤ q)
      - Upper tail: λ_U(q) = P(U₂ ≥ q | U₁ ≥ q)

    where U₁ and U₂ are the pseudo-observations.

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        A 2D array with exactly 2 columns representing the pseudo-observations.

    quantiles : array-like of shape (n_quantiles,)
        A 1D array of quantile levels (values between 0 and 1) at which to compute the
        concentration.

    Returns
    -------
    concentration : ndarray of shape (n_quantiles,)
        An array of empirical tail concentration values for the given quantiles.

    References
    ----------
    .. [1] "Quantitative Risk Management: Concepts, Techniques, and Tools",
        McNeil, Frey, Embrechts (2005)

    Raises
    ------
    ValueError
        If X is not a 2D array with exactly 2 columns or if quantiles are not in [0, 1].
    """
    X = np.asarray(X)
    if X.ndim != 2 or X.shape[1] != 2:
        raise ValueError("X must be a 2D array with exactly 2 columns.")
    if not np.all((X >= 0) & (X <= 1)):
        raise ValueError("X must be pseudo-observation in the interval `[0, 1]`")
    quantiles = np.asarray(quantiles)
    if not np.all((quantiles >= 0) & (quantiles <= 1)):
        raise ValueError("quantiles must be between 0.0 and 1.0.")

    def func(q: np.ndarray, is_lower: bool) -> np.ndarray:
        op = operator.le if is_lower else operator.ge
        cond = op(X[:, 0, np.newaxis], q)
        count = np.count_nonzero(cond, axis=0).astype(float)
        mask = count != 0
        count[mask] = (
            np.count_nonzero(cond & op(X[:, 1, np.newaxis], q), axis=0)[mask]
            / count[mask]
        )
        return count

    concentration = np.where(
        quantiles <= 0.5, func(quantiles, True), func(quantiles, False)
    )
    return concentration


def plot_tail_concentration(
    tail_concentration_dict: dict[str, npt.ArrayLike],
    quantiles: np.ndarray,
    title: str = "Empirical Tail Dependencies",
    smoothing: float | None = 0.5,
) -> go.Figure:
    """
    Plot the empirical tail concentration curves.

    This function takes a dictionary where keys are dataset names and values are the
    corresponding tail concentration arrays computed at the given quantiles. It then
    creates a Plotly figure with the tail concentration curves. The x-axis (quantiles)
    and y-axis (tail concentration) are both formatted as percentages.

    Parameters
    ----------
    tail_concentration_dict : dict[str, ArrayLike]
        A dictionary mapping dataset names to their tail concentration values.

    quantiles : array-like of shape (n_quantiles,)
        The quantile levels at which the tail concentration has been computed.

    title : str, default="Empirical Tail Dependencies"
        The title for the plot.

    smoothing : float or None, default=0.5
        Smoothing parameter for the spline line shape. If provided, the curves will be
        smoothed using a spline interpolation.

    Returns
    -------
    fig : go.Figure
        A Plotly figure object containing the tail concentration curves.

    Raises
    ------
    ValueError
        If the smoothing parameter is not in the allowed range.
    """
    if smoothing is not None and not (0 <= smoothing <= 1.3):
        raise ValueError("The smoothing parameter must be between 0 and 1.3.")

    quantiles = np.asarray(quantiles)
    traces = []
    # Determine the line shape and include the smoothing parameter if applicable.
    if smoothing is not None:
        line_dict = {"shape": "spline", "smoothing": smoothing}
    else:
        line_dict = {"shape": "linear"}

    # Iterate over each dataset name and its corresponding data array.
    for name, concentration in tail_concentration_dict.items():
        concentration = np.asarray(concentration)
        trace = go.Scatter(
            x=quantiles,
            y=np.asarray(concentration),
            mode="lines",
            name=name,
            line=line_dict,
        )
        traces.append(trace)

    # Create the Plotly figure.
    fig = go.Figure(data=traces)
    fig.update_layout(
        title=title,
        xaxis_title="Quantile",
        yaxis_title="Tail Concentration",
    )
    # Update both axes to show percentages with an enhanced grid.
    fig.update_xaxes(
        range=[0, 1],
        tickformat=".0%",
        dtick=0.05,
        showgrid=True,
        gridwidth=1,
        gridcolor="lightgrey",
    )
    fig.update_yaxes(
        range=[0, 1],
        tickformat=".0%",
        dtick=0.05,
        showgrid=True,
        gridwidth=1,
        gridcolor="lightgrey",
    )

    return fig


def _select_rotation_itau(
    func: Callable, X: np.ndarray, theta: float
) -> CopulaRotation:
    """
    Select the optimal copula rotation based on a provided function.

    This helper function applies each rotation defined in CopulaRotation to the data X,
    computes a criterion value using the provided function (which takes X and theta as
    arguments), and returns the rotation that minimizes this value.

    Parameters
    ----------
    func : Callable
       A function that computes a criterion (e.g., a negative log-likelihood) for a given
       rotated dataset and copula parameter theta.

    X : ndarray of shape (n_observations, 2)
       A 2D array of bivariate inputs.

    theta : float
       The copula parameter to be used in the criterion function.

    Returns
    -------
    CopulaRotation
       The rotation (an element of CopulaRotation) that minimizes the criterion value.
    """
    results = {}
    for rotation in CopulaRotation:
        X_rotated = _apply_copula_rotation(X, rotation=rotation)
        results[rotation] = func(X=X_rotated, theta=theta)
    best_rotation = min(results, key=results.get)
    return best_rotation


def _select_theta_and_rotation_mle(
    func: Callable, X: np.ndarray, bounds: tuple[float, float], tolerance: float = 1e-4
) -> tuple[float, CopulaRotation]:
    """
    Select the optimal copula parameter theta and rotation using maximum likelihood
    estimation.

    For each rotation defined in CopulaRotation, this function applies the rotation to
    X, then minimizes the negative log-likelihood over theta using a bounded scalar
    optimization. It returns the theta and rotation that yield the minimum criterion
    value.

    Parameters
    ----------
    func : Callable
        A function that computes the negative log-likelihood (or similar criterion) for a
        given value of theta and rotated data X.

    X : ndarray of shape (n_observations, 2)
        A 2D array of bivariate inputs.

    bounds : tuple[float, float]
        The lower and upper bounds for the copula parameter theta.

    tolerance : float, default=1e-4
        The tolerance for the scalar minimization optimization.

    Returns
    -------
    tuple
        A tuple (theta, rotation) where theta is the optimal copula parameter and
        rotation is the corresponding CopulaRotation.

    Raises
    ------
    RuntimeError
        If the optimization fails for all rotations.
    """
    results = []
    for rotation in CopulaRotation:
        X_rotated = _apply_copula_rotation(X, rotation=rotation)
        result = so.minimize_scalar(
            func,
            args=(X_rotated,),
            bounds=bounds,
            method="bounded",
            options={"xatol": tolerance},
        )
        if result.success:
            results.append(
                {
                    "neg_log_likelihood": result.fun,
                    "theta": result.x,
                    "rotation": rotation,
                }
            )
        else:
            warnings.warn(
                f"Optimization failed for rotation {rotation}: {result.message}",
                RuntimeWarning,
                stacklevel=2,
            )
    if len(results) == 0:
        raise RuntimeError("Optimization failed for all rotations")

    best = min(results, key=lambda d: d["neg_log_likelihood"])
    return best["theta"], best["rotation"]


def _apply_copula_rotation(X: npt.ArrayLike, rotation: CopulaRotation) -> np.ndarray:
    r"""Apply a bivariate copula rotation using the standard (clockwise) convention.

    The transformations are defined as follows:

    - `CopulaRotation.R0` (0°): :math:`(u, v) \mapsto (u, v)`
    - `CopulaRotation.R90` (90°): :math:`(u, v) \mapsto (v,\, 1 - u)`
    - `CopulaRotation.R180` (180°): :math:`(u, v) \mapsto (1 - u,\, 1 - v)`
    - `CopulaRotation.R270` (270°): :math:`(u, v) \mapsto (1 - v,\, u)`

    Parameters
    ----------
    X : array-like of shape (n_observations, 2)
        An array of bivariate inputs `(u, v)` where each row represents a
        bivariate observation.

    rotation : CopulaRotation
        The rotation to apply to the copula (default is no rotation).

    Returns
    -------
    rotated_X: ndarray of shape (n_observations, 2)
        The rotated data array.
    """
    match rotation:
        case CopulaRotation.R0:
            # No rotation
            pass
        case CopulaRotation.R90:
            # (u, v) -> (v, 1 - u)
            X = np.column_stack([X[:, 1], 1.0 - X[:, 0]])
        case CopulaRotation.R180:
            # (u, v) -> (1 - u, 1 - v)
            X = 1.0 - X
        case CopulaRotation.R270:
            # (u, v) -> (1 - v, u)
            X = np.column_stack([1.0 - X[:, 1], X[:, 0]])
        case _:
            raise ValueError(f"Unsupported rotation: {rotation}")
    return X


def _apply_margin_swap(X: np.ndarray, first_margin: bool) -> np.ndarray:
    """
    Swap the columns of X if first_margin is False.

    If first_margin is True, X is returned unchanged; otherwise, the columns
    of X are swapped.

    Parameters
    ----------
    X : ndarray of shape (n_observations, 2)
        A 2D array of bivariate inputs (u, v).
    first_margin : bool
        If True, no swap is performed; if False, the columns of X are swapped.

    Returns
    -------
    X_swapped : ndarray of shape (n_observations, 2)
        The data array with columns swapped if first_margin is False.
    """
    assert X.ndim == 2
    assert X.shape[1] == 2
    if first_margin:
        return X[:, [1, 0]]
    return X


def _apply_rotation_cdf(
    func: Callable, X: np.ndarray, rotation: CopulaRotation, **kwargs
) -> np.ndarray:
    """
    Apply a copula rotation to X and compute the corresponding CDF values.

    Parameters
    ----------
    func : Callable
       A function that computes the CDF given data X and additional keyword arguments.

    X : ndarray of shape (n_observations, 2)
       A 2D array of bivariate inputs.

    rotation : CopulaRotation
       The rotation to apply.

    **kwargs
       Additional keyword arguments to pass to the CDF function.

    Returns
    -------
    rotated_cdf : ndarray of shape (n_observations,)
       The transformed CDF values after applying the rotation.
    """
    rotated_X = _apply_copula_rotation(X, rotation=rotation)
    cdf = func(X=rotated_X, **kwargs)

    match rotation:
        case CopulaRotation.R0:
            pass
        case CopulaRotation.R90:
            cdf = X[:, 1] - cdf
        case CopulaRotation.R180:
            cdf = np.sum(X, axis=1) - 1 + cdf
        case CopulaRotation.R270:
            cdf = X[:, 0] - cdf
        case _:
            raise ValueError(f"Unsupported rotation: {rotation}")

    return cdf


def _apply_rotation_partial_derivatives(
    func: Callable,
    X: np.ndarray,
    rotation: CopulaRotation,
    first_margin: bool,
    **kwargs,
) -> np.ndarray:
    """
    Apply a copula rotation to X and compute the corresponding partial derivatives.

    This function rotates the data X using the specified rotation and then computes
    the partial derivative (h-function) using the provided function. The result is then
    adjusted according to the rotation and the margin of interest.

    Parameters
    ----------
    func : Callable
        A function that computes the partial derivative (h-function) given X, the
        margin, and any additional keyword arguments.

    X : ndarray of shape (n_observations, 2)
        A 2D array of bivariate inputs.

    rotation : CopulaRotation
        The rotation to apply.

    first_margin : bool
        If True, compute the partial derivative with respect to the first margin;
        otherwise, compute it with respect to the second margin.

    **kwargs
        Additional keyword arguments to pass to the partial derivative function.

    Returns
    -------
    z : ndarray of shape (n_observations,)
        The transformed partial derivative values after applying the rotation.
    """
    rotated_X = _apply_copula_rotation(X, rotation=rotation)

    match rotation:
        case CopulaRotation.R0:
            z = func(X=rotated_X, first_margin=first_margin, **kwargs)
        case CopulaRotation.R90:
            if first_margin:
                z = func(X=rotated_X, first_margin=not first_margin, **kwargs)
            else:
                z = 1 - func(X=rotated_X, first_margin=not first_margin, **kwargs)
        case CopulaRotation.R180:
            z = 1 - func(X=rotated_X, first_margin=first_margin, **kwargs)
        case CopulaRotation.R270:
            if first_margin:
                z = 1 - func(X=rotated_X, first_margin=not first_margin, **kwargs)
            else:
                z = func(X=rotated_X, first_margin=not first_margin, **kwargs)
        case _:
            raise ValueError(f"Unsupported rotation: {rotation}")
    return z



"""Vine Copula Estimation."""

# Copyright (c) 2025
# Author: Hugo Delatte <delatte.hugo@gmail.com>
# Credits: Matteo Manzi, Vincent Maladière, Carlo Nicolini
# SPDX-License-Identifier: BSD-3-Clause

# This is a novel implementation of R-vine that is fully OOP based on graph theory with
# new features to also capture C-like and clustered dependency structures.
#
# Unlike previous implementations that relied on a procedural paradigm, this version
# adopts an object-oriented design for constructing vine copulas.
#
# The validity of a vine tree is based on the following principle:
#   - In tree T_{k-1}, each edge represents a subset of k variables.
#   - Two such edges can form a new edge in tree T_k if and only if they share exactly
#     k-1 variables.
#
# In an OOP approach, we can use the below equivalent graphical adjacency rule:
# Two edges in T_{k-1} (which are “nodes” when building T_k) are connected by an edge
# in T_k if and only if they share exactly one node in T_{k-1}.
#
#  By using the concept of central assets in the MST, this novel implementation is able
#  to capture clustered or C-like dependency structures, allowing for more nuanced
#  representation of hierarchical relationships among assets and improving conditional
#  sampling.

import contextlib
import numbers
import warnings
from collections import deque

import numpy as np
import numpy.typing as npt
import plotly.express as px
import plotly.graph_objects as go
import sklearn.utils as sku
import sklearn.utils.parallel as skp
import sklearn.utils.validation as skv

from skfolio.distribution._base import SelectionCriterion
from skfolio.distribution.copula import (
    UNIFORM_MARGINAL_EPSILON,
    BaseBivariateCopula,
    ClaytonCopula,
    GaussianCopula,
    GumbelCopula,
    JoeCopula,
    StudentTCopula,
    select_bivariate_copula,
)
from skfolio.distribution.multivariate._base import BaseMultivariateDist
from skfolio.distribution.multivariate._utils import (
    ChildNode,
    DependenceMethod,
    Edge,
    RootNode,
    Tree,
)
from skfolio.distribution.univariate import (
    BaseUnivariateDist,
    Gaussian,
    JohnsonSU,
    StudentT,
    select_univariate_dist,
)
from skfolio.utils.figure import kde_trace
from skfolio.utils.tools import input_to_array, validate_input_list

_UNIFORM_SAMPLE_EPSILON = 1e-14


class VineCopula(BaseMultivariateDist):
    """
    Regular Vine Copula Estimator.

    This model first fits the best univariate distribution for each asset, transforming
    the data to uniform marginals via the fitted CDFs. Then, it constructs a regular
    vine copula by sequentially selecting the best bivariate copula from a list of
    candidates for each edge in the vine using a maximum spanning tree algorithm based
    on a given dependence measure [1]_.

    Regular vines captures complex, fat-tailed dependencies and tail co-movements
    between asset returns.

    It also supports conditional sampling, enabling stress testing and scenario analysis
    by generating samples under specified conditions.

    Moreover, by marking some assets as central, this novel implementation is able to
    capture clustered or C-like dependency structures, allowing for more nuanced
    representation of hierarchical relationships among assets and improving conditional
    sampling and stress testing.

    Parameters
    ----------
    fit_marginals : bool, default=True
        Whether to fit marginal distributions to each asset before constructing the
        vine. If True, the data will be transformed to uniform marginals using the
        fitted CDFs.

    marginal_candidates : list[BaseUnivariateDist], optional
        Candidate univariate distribution estimators to fit the marginals.
        If None, defaults to `[Gaussian(), StudentT(), JohnsonSU()]`.

    copula_candidates : list[BaseBivariateCopula], optional
        Candidate bivariate copula estimators. If None, defaults to
        `[GaussianCopula(), StudentTCopula(), ClaytonCopula(), GumbelCopula(), JoeCopula()]`.

    max_depth : int or None, default=4
        Maximum vine depth (truncated level). Must be greater than 1.
        `None` means that no truncation is applied. The default is 4.

    log_transform : bool | dict[str, bool] | array-like of shape (n_assets, ), default=False
        If True, the simple returns provided as input will be transformed to log returns
        before fitting the vine copula. That is, each return R is transformed via
        r = log(1+R). After sampling, the generated log returns are converted back to
        simple returns using R = exp(r) - 1.

        If a boolean is provided, it is applied to each asset.
        If a dictionary is provided, its (key/value) pair must be the
        (asset name/boolean) and the input `X` of the `fit` method must be a
        DataFrame with the assets names in columns.

    central_assets : array-like of asset names or asset positions, optional
        Assets that should be centrally placed during vine construction.
        If None, no asset is forced to the center.
        If an array-like of **integer** is provided, its values must be asset positions.
        If an array-like of **string** is provided, its values must be asset names and
        the input `X` of the `fit` method must be a DataFrame with the assets names in
        columns. Assets marked as central are forced to occupy central positions in the
        vine, leading to C-like or clustered structure. This is needed for conditional
        sampling, where the conditioning assets should be central nodes.

        For example:

        - If only asset 1 is marked as central, it will be connected to all other
           assets in the first tree (yielding a C-like structure for the initial
           tree), with subsequent trees following the standard R-vine pattern.
        - If asset 1 and asset 2 are marked as central, they will be connected
           together and the remaining assets will connect to either asset 1 or asset
           2 (forming a clustered structure in the initial trees). In the next tree,
           the edge between asset 1 and asset 2 becomes the central node, with
           subsequent trees following the standard R-vine structure.
        - This logic extends naturally to more than two central assets.

    dependence_method : DependenceMethod, default=DependenceMethod.KENDALL_TAU
        The dependence measure used to compute edge weights for the MST.
        Possible values are:

        - KENDALL_TAU
        - MUTUAL_INFORMATION
        - WASSERSTEIN_DISTANCE

    selection_criterion : SelectionCriterion, default=SelectionCriterion.AIC
        The criterion used for univariate and copula selection. Possible values are:

        - SelectionCriterion.AIC : Akaike Information Criterion
        - SelectionCriterion.BIC : Bayesian Information Criterion

    independence_level : float, default=0.05
        Significance level used for the Kendall tau independence test during copula
        fitting. If the p-value exceeds this threshold, the null hypothesis of
        independence is accepted, and the pair copula is modeled using the
        `IndependentCopula()` class.

    n_jobs : int, optional
        The number of jobs to run in parallel for `fit` of all `estimators`.
        The value `-1` means using all processors.
        The default (`None`) means 1 unless in a `joblib.parallel_backend` context.

    random_state : int, RandomState instance or None, default=None
        Seed or random state to ensure reproducibility.

    Attributes
    ----------
    trees_ : list[Tree]
        List of constructed vine trees.

    marginal_distributions_ : list[BaseUnivariateDist]
        List of fitted marginal distributions (if fit_marginals is True).

    n_features_in_ : int
        Number of assets seen during `fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of assets seen during `fit`. Defined only when `X`
        has assets names that are all strings.

    Examples
    --------
    >>> from skfolio.datasets import load_factors_dataset
    >>> from skfolio.preprocessing import prices_to_returns
    >>> from skfolio.distribution import VineCopula
    >>>
    >>> # Load historical prices and convert them to returns
    >>> prices = load_factors_dataset()
    >>> X = prices_to_returns(prices)
    >>>
    >>> # Instanciate the VineCopula model
    >>> vine = VineCopula()
    >>> # Fit the model
    >>> vine.fit(X)
    >>> # Display the vine trees and fitted copulas
    >>> vine.display_vine()
    >>> # Log-likelihood, AIC and BIC
    >>> vine.score(X)
    >>> vine.aic(X)
    >>> vine.bic(X)
    >>>
    >>> # Generate 10 samples from the fitted vine copula
    >>> samples = vine.sample(n_samples=10)
    >>>
    >>> # Set QUAL, SIZE and MTUM as central
    >>> vine = VineCopula(central_assets=["QUAL", "SIZE", "MTUM"])
    >>> vine.fit(X)
    >>> # Sample by conditioning on QUAL and SIZE returns
    >>> samples = vine.sample(
    ...    n_samples=4,
    ...    conditioning={
    ...        "QUAL": [-0.1, -0.2, -0.3, -0.4],
    ...        "SIZE": -0.2,
    ...        "MTUM": (None, -0.3) # MTUM sampled between -Inf and -30%
    ...    },
    ...)
    >>> # Plots Scatter matrix of sampled returns vs historical X
    >>> fig = vine.plot_scatter_matrix(X=X)
    >>> fig.show()
    >>>
    >>> # Plots univariate distributions of sampled returns vs historical X
    >>> fig = vine.plot_marginal_distributions(X=X)
    >>> fig.show()

    References
    ----------
    .. [1] "Selecting and estimating regular vine copulae and application to financial
        returns" Dißmann, Brechmann, Czado, and Kurowicka (2013).

    .. [2] "Growing simplified vine copula trees: improving Dißmann's algorithm"
        Krausa and Czado (2017).

    .. [3] "Pair-copula constructions of multiple dependence" Aas, Czado, Frigessi,
        Bakken (2009).

    .. [4] "Pair-Copula Constructions for Financial Applications: A Review"
        Aas and Czado (2016).

    .. [5] "Conditional copula simulation for systemic risk stress testing"
        Brechmann, Hendrich, Czado (2013)
    """

    trees_: list[Tree]
    marginal_distributions_: list[BaseUnivariateDist]
    n_features_in_: int
    feature_names_in_: np.ndarray
    central_assets_: set[int | str]

    _log_transform: np.ndarray

    def __init__(
        self,
        fit_marginals: bool = True,
        marginal_candidates: list[BaseUnivariateDist] | None = None,
        copula_candidates: list[BaseBivariateCopula] | None = None,
        max_depth: int | None = 4,
        log_transform: bool | dict[str, bool] | npt.ArrayLike = False,
        central_assets: list[int | str] | None = None,
        dependence_method: DependenceMethod = DependenceMethod.KENDALL_TAU,
        selection_criterion: SelectionCriterion = SelectionCriterion.AIC,
        independence_level: float = 0.05,
        n_jobs: int | None = None,
        random_state: int | None = None,
    ):
        super().__init__(random_state=random_state)
        self.fit_marginals = fit_marginals
        self.marginal_candidates = marginal_candidates
        self.copula_candidates = copula_candidates
        self.max_depth = max_depth
        self.log_transform = log_transform
        self.central_assets = central_assets
        self.dependence_method = dependence_method
        self.selection_criterion = selection_criterion
        self.independence_level = independence_level
        self.n_jobs = n_jobs

    @property
    def n_params(self) -> int:
        """Number of model parameters."""
        skv.check_is_fitted(self)
        k = 0
        if self.fit_marginals:
            k = sum([dist.n_params for dist in self.marginal_distributions_])

        for tree in self.trees_:
            for edge in tree.edges:
                k += edge.copula.n_params

        return k

    def fit(self, X: npt.ArrayLike, y=None) -> "VineCopula":
        """
        Fit the Vine Copula model to the data.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        y : Ignored
            Not used, present for API consistency by convention.

        Returns
        -------
        self : VineCopula
            The fitted VineCopula instance.

        Raises
        ------
        ValueError
            If the number of assets is less than or equal to 2, or if max_depth <= 1.
        """
        X = skv.validate_data(self, X, dtype=np.float64)
        _, n_assets = X.shape

        if n_assets <= 2:
            raise ValueError(
                f"The number of assets must be higher than 2, got {n_assets}"
            )
        if self.max_depth is not None and self.max_depth <= 1:
            raise ValueError(f"`max_depth` must be higher than 1, got {self.max_depth}")

        if np.isscalar(self.log_transform):
            self._log_transform = np.array([self.log_transform] * n_assets, dtype=bool)
        else:
            self._log_transform = input_to_array(
                items=self.log_transform,
                n_assets=n_assets,
                fill_value=False,
                dim=1,
                assets_names=getattr(self, "feature_names_in_", None),
                name="log_transform",
            )

        if np.any(self._log_transform):
            X = np.where(self._log_transform, np.log1p(X), X)

        depth = n_assets - 1
        if self.max_depth is not None:
            depth = min(self.max_depth, depth)

        self.central_assets_ = set()
        if self.central_assets is not None:
            self.central_assets_ = set(
                validate_input_list(
                    items=self.central_assets,
                    n_assets=n_assets,
                    assets_names=getattr(self, "feature_names_in_", None),
                    name="central_assets",
                    raise_if_string_missing=False,
                )
            )

        marginal_candidates = (
            self.marginal_candidates
            if self.marginal_candidates is not None
            else [
                Gaussian(),
                StudentT(),
                JohnsonSU(),
            ]
        )

        copula_candidates = (
            self.copula_candidates
            if self.copula_candidates is not None
            else [
                GaussianCopula(),
                StudentTCopula(),
                ClaytonCopula(),
                GumbelCopula(),
                JoeCopula(),
            ]
        )

        if self.fit_marginals:
            # Fit marginal distributions
            self.marginal_distributions_ = skp.Parallel(n_jobs=self.n_jobs)(
                skp.delayed(select_univariate_dist)(
                    X=X[:, [i]],
                    distribution_candidates=marginal_candidates,
                    selection_criterion=self.selection_criterion,
                )
                for i in range(n_assets)
            )
            # Transform data to uniform marginals using the fitted CDF
            X = np.hstack(
                [
                    dist.cdf(X[:, [i]])
                    for i, dist in enumerate(self.marginal_distributions_)
                ]
            )

        # Ensure values are within [0, 1].
        if not np.all((X >= 0) & (X <= 1)):
            raise ValueError(
                "If `fit_marginals` is set to False, X must be in the interval "
                "`[0, 1]`, typically obtained via marginal CDF transformation."
            )
        # Handle potential numerical issues by ensuring X doesn't contain exact 0 or 1.
        X = np.clip(X, UNIFORM_MARGINAL_EPSILON, 1 - UNIFORM_MARGINAL_EPSILON)

        trees: list[Tree] = []
        for level in range(depth):
            if level == 0:
                # Initial tree: each asset is a node.
                tree = Tree(
                    level=level,
                    nodes=[
                        RootNode(
                            ref=i,
                            pseudo_values=X[:, i],
                            central=i in self.central_assets_,
                        )
                        for i in range(n_assets)
                    ],
                )
            else:
                # Subsequent trees: nodes represent edges from the previous tree.
                tree = Tree(
                    level=level, nodes=[ChildNode(ref=edge) for edge in trees[-1].edges]
                )
            # Set edges via Maximum Spanning Tree using the specified dependence method.
            tree.set_edges_from_mst(dependence_method=self.dependence_method)
            assert len(tree.edges) == n_assets - level - 1

            # Fit bivariate copulas for each edge.
            copulas = skp.Parallel(n_jobs=self.n_jobs)(
                skp.delayed(select_bivariate_copula)(
                    X=edge.get_X(),
                    copula_candidates=copula_candidates,
                    selection_criterion=self.selection_criterion,
                    independence_level=self.independence_level,
                )
                for edge in tree.edges
            )
            for edge, copula in zip(tree.edges, copulas, strict=True):
                edge.copula = copula

            # Clear previous tree cache to free memory.
            if level > 0:
                trees[-1].clear_cache()

            trees.append(tree)

        # Clear last tree
        trees[-1].clear_cache()

        # Attach a node to each terminal edge (used for sampling).
        for edge in trees[-1].edges:
            ChildNode(ref=edge)

        self.trees_ = trees
        return self

    def score_samples(self, X: npt.ArrayLike) -> np.ndarray:
        r"""Compute the log-likelihood of each sample (log-pdf) under the model.

        Parameters
        ----------
        X : array-like of shape (n_observations, n_assets)
            Price returns of the assets.

        Returns
        -------
        density : ndarray of shape (n_observations,)
            The log-likelihood of each sample under the fitted vine copula.
        """
        skv.check_is_fitted(self)
        X = skv.validate_data(self, X, dtype=np.float64, reset=False)
        self.clear_cache()

        if np.any(self._log_transform):
            X = np.where(self._log_transform, np.log1p(X), X)

        if not self.fit_marginals:
            score_samples = np.zeros(len(X))
        else:
            score_samples = np.sum(
                [
                    dist.score_samples(X[:, [i]])
                    for i, dist in enumerate(self.marginal_distributions_)
                ],
                axis=0,
            )
            X = np.hstack(
                [
                    dist.cdf(X[:, [i]])
                    for i, dist in enumerate(self.marginal_distributions_)
                ]
            )

        # Handle potential numerical issues by ensuring X doesn't contain exact 0 or 1.
        X = np.clip(X, UNIFORM_MARGINAL_EPSILON, 1 - UNIFORM_MARGINAL_EPSILON)

        for i, node in enumerate(self.trees_[0].nodes):
            node.pseudo_values = X[:, i]

        for tree in self.trees_:
            for edge in tree.edges:
                score_samples += edge.copula.score_samples(X=edge.get_X())
        self.clear_cache()

        return score_samples

    def sample(
        self,
        n_samples: int = 1,
        conditioning: dict[int | str : float | tuple[float, float] | npt.ArrayLike]
        | None = None,
    ) -> np.ndarray:
        """Generate random samples from the vine copula.

        This method generates `n_samples` from the fitted vine copula model. The
        resulting samples represent multivariate observations drawn according to the
        dependence structure captured by the vine copula.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to generate.

        conditioning : dict[int | str, float | tuple[float, float] | array-like], optional
            A dictionary specifying conditioning information for one or more assets.
            The dictionary keys are asset indices or names, and the values define how
            the samples are conditioned for that asset. Three types of conditioning
            values are supported:

            1. **Fixed value (float):**
               If a float is provided, all samples are generated under the condition
               that the asset takes exactly that value.

            2. **Bounds (tuple of two floats):**
               If a tuple `(min_value, max_value)` is provided, samples are generated
               under the condition that the asset's value falls within the specified
               bounds. Use `-np.Inf` for no lower bound or `np.Inf` for no upper bound.

            3. **Array-like (1D array):**
               If an array-like of length `n_samples` is provided, each sample is
               conditioned on the corresponding value in the array for that asset.

            **Important:** When using conditional sampling, it is recommended that the
            assets you condition on are set as central during the vine copula
            construction. This can be specified via the `central_assets` parameter in
            the vine copula instantiation.

        Returns
        -------
        X : array-like of shape (n_samples, n_assets)
            A two-dimensional array where each row is a multivariate observation sampled
            from the vine copula.
        """
        skv.check_is_fitted(self)
        self.clear_cache()
        n_assets = self.n_features_in_

        rng, conditioning_vars, conditioning_clean, uniform_cond_samples = (
            self._init_conditioning(n_samples=n_samples, conditioning=conditioning)
        )

        # Determine sampling order based on vine structure.
        sampling_order = self._sampling_order(conditioning_vars=conditioning_vars)

        # Generate independent Uniform(0,1) samples for non-conditioned nodes.
        X_rand = rng.random(size=(n_assets - len(conditioning_vars), n_samples))

        # Propagate samples through the vine structure bottom-up.

        # We perform a first pass by only recording the number of Node visits in
        # each Node without propagating any data. This count is used for optimally
        # clearing cache during sampling.
        with self._count_node_visits():
            _propagate_samples(
                X_rand,
                sampling_order,
                conditioning_vars,
                uniform_cond_samples,
            )
        # We now perform the second pass with full data propagation.
        _propagate_samples(
            X_rand,
            sampling_order,
            conditioning_vars,
            uniform_cond_samples,
        )

        # Collect samples from the root tree.
        samples = np.stack([node.pseudo_values for node in self.trees_[0].nodes]).T
        self.clear_cache()

        # Avoid Inf
        samples = np.clip(
            samples,
            a_min=_UNIFORM_SAMPLE_EPSILON,
            a_max=1 - _UNIFORM_SAMPLE_EPSILON,
        )

        # Transform samples back to the original scale using inverse CDF
        # (if marginals were fitted).
        if self.fit_marginals:
            samples = np.hstack(
                [
                    dist.ppf(samples[:, [i]])
                    for i, dist in enumerate(self.marginal_distributions_)
                ]
            )

        # Reverse the log-return transformation if log_transform is True.
        if np.any(self._log_transform):
            # A known effect when sampling with log transform is that large sampling
            # log returns explode after applying the reverse log transform into
            # unrealistic simple returns.
            # A common approach is to apply Box-Cox transform above some threshold.
            # For financial returns, log returns above 200% are transformed linearly
            # instead of exponentially.
            box_cox_threshold = np.log(3)
            samples[:, self._log_transform] = np.where(
                samples[:, self._log_transform] <= box_cox_threshold,
                np.exp(samples[:, self._log_transform]) - 1,
                np.exp(box_cox_threshold)
                + (samples[:, self._log_transform] - box_cox_threshold)
                - 1,
            )

        # To avoid Inf values and numerical instability, conditional values converted to
        # uniforms are clipped to [1e-14 , 1-1e-14]. This means that if the conditional
        # value has an extremely low probability, its final value will be bounded by
        # [ppf(1e-14) , ppf(1-1e-14)]. To keep conditional values accurate even for
        # extremely low probability, we force them back in the final samples.
        for var, cond in conditioning_clean.items():
            if isinstance(cond, tuple):
                samples[:, var] = np.clip(samples[:, var], a_min=cond[0], a_max=cond[1])
            else:
                samples[:, var] = cond

        return samples

    def clear_cache(self, clear_count: bool = True):
        """Clear cached intermediate results in the vine trees."""
        for tree in self.trees_:
            tree.clear_cache(clear_count=clear_count)
        for edge in self.trees_[-1].edges:
            edge.ref_node.clear_cache(clear_count=clear_count)

    def _conditioning_count(self) -> np.ndarray:
        """Compute cumulative counts of conditioning set occurrences in the vine.

        Returns
        -------
        conditioning_counts : ndarray of shape (n_trees, n_assets)
            Array of cumulative conditioning counts.
        """
        n_assets = self.n_features_in_

        conditioning_counts = np.cumsum(
            [
                np.bincount(
                    [s for edge in tree.edges for s in edge.cond_sets.conditioning],
                    minlength=n_assets,
                )
                for tree in self.trees_
            ],
            axis=0,
        )

        return conditioning_counts

    def _init_conditioning(
        self,
        n_samples: int,
        conditioning: dict[int | str : float | tuple[float, float] | npt.ArrayLike],
    ) -> tuple[
        np.random.RandomState, set[int], dict[int, float], dict[int, np.ndarray]
    ]:
        """
        Initialised conditioning variables used in the conditioning sampling.

        Parameters
        ----------
        n_samples : int, default=1
            Number of samples to generate.

        conditioning : dict[int | str, float | tuple[float, float] | array-like], optional
            A dictionary specifying conditioning information for one or more assets.
            The dictionary keys are asset indices or names, and the values define how
            the samples are conditioned for that asset. Three types of conditioning
            values are supported:

            1. **Fixed value (float):**
               If a float is provided, all samples are generated under the condition
               that the asset takes exactly that value.

            2. **Bounds (tuple of two floats):**
               If a tuple `(min_value, max_value)` is provided, samples are generated
               under the condition that the asset's value falls within the specified
               bounds. Use `-np.Inf` for no lower bound or `np.Inf` for no upper bound.

            3. **Array-like (1D array):**
               If an array-like of length `n_samples` is provided, each sample is
               conditioned on the corresponding value in the array for that asset.

        Returns
        -------
        random_sate : RandomState
            Numpy Random State.

        conditioning_vars : set[int]
            The conditioning variables.

        conditioning_clean : dict[int, float]
            The cleaned conditioning dictionary.

        uniform_cond_samples : dict[int, np.ndarray]
            The uniform conditioning samples corresponding to the `conditioning`.
        """
        rng = sku.check_random_state(self.random_state)

        conditioning_vars = set()
        conditioning_clean = dict()
        uniform_cond_samples = dict()

        if conditioning is None:
            return rng, conditioning_vars, conditioning_clean, uniform_cond_samples

        if not isinstance(conditioning, dict):
            raise ValueError(
                "When provided, `conditioning` must be a dictionary mapping each asset "
                "to its corresponding conditioning value, "
                f"received {type(conditioning)}"
            )

        n_assets = self.n_features_in_

        conditioning_vars = validate_input_list(
            items=list(conditioning.keys()),
            n_assets=n_assets,
            assets_names=getattr(self, "feature_names_in_", None),
            name="conditioning",
        )

        missing_central_vars = set(conditioning_vars).difference(self.central_assets_)

        if not set(conditioning_vars).issubset(set(range(n_assets))):
            raise ValueError(
                "The keys of `conditioning` must be asset indices or names "
                "from the input X."
            )
        if len(conditioning_vars) >= n_assets:
            raise ValueError(
                "`conditioning` must be provided for strictly fewer assets "
                "than the total."
            )

        if missing_central_vars:
            warnings.warn(
                "When performing conditional sampling, it is recommended to set "
                "conditioning assets as central during Vine Copula construction. "
                "The following conditioning assets were not set as central: "
                f"{missing_central_vars}. "
                "This can be achieved by using the `central_assets` parameter.",
                stacklevel=2,
            )

        for var, value in zip(conditioning_vars, conditioning.values(), strict=True):
            if isinstance(value, tuple):
                if len(value) != 2:
                    raise ValueError(
                        "When a tuple is provided in `conditioning`, it must be"
                        "of length 2, representing the conditioning bounds: "
                        "(min_value, max_value)"
                    )
                min_value, max_value = value
                if min_value is None:
                    min_value = -np.inf
                if max_value is None:
                    max_value = np.inf

                if min_value >= max_value:
                    raise ValueError(
                        "The conditioning tuple lower bound must be lower than "
                        "its upper bound."
                    )
                conditioning_clean[var] = (min_value, max_value)

                if self._log_transform[var]:
                    if not np.isneginf(min_value):
                        min_value = np.log1p(min_value)
                    if not np.isposinf(max_value):
                        max_value = np.log1p(max_value)

                if self.fit_marginals:
                    # Transform the bounds using the marginal CDF
                    dist = self.marginal_distributions_[var]
                    u_min = 0.0 if np.isneginf(min_value) else dist.cdf(min_value)
                    u_max = 1.0 if np.isposinf(max_value) else dist.cdf(max_value)
                else:
                    u_min, u_max = min_value, max_value

                # Sample uniformly in the transformed interval.
                samples = rng.uniform(low=u_min, high=u_max, size=n_samples)

            elif np.isscalar(value):
                if not isinstance(value, numbers.Number):
                    raise ValueError(
                        f"Conditioning values should be numbers, got {value}"
                    )
                conditioning_clean[var] = value
                if self._log_transform[var]:
                    value = np.log1p(value)
                if self.fit_marginals:
                    # Transform conditioning value using the fitted marginal CDF.
                    value = self.marginal_distributions_[var].cdf(value)
                samples = np.full(n_samples, value)
            else:
                samples = np.asarray(value)
                conditioning_clean[var] = samples
                if samples.ndim != 1 or samples.shape[0] != n_samples:
                    raise ValueError(
                        "When an array is provided in `conditioning`, it must be a "
                        f"1D array of length n_samples={n_samples}, got {samples.ndim}D of "
                        f"length {samples.shape[0]}"
                    )
                if self._log_transform[var]:
                    samples = np.log1p(samples)
                if self.fit_marginals:
                    # Transform conditioning samples using the fitted marginal CDF.
                    samples = self.marginal_distributions_[var].cdf(samples)
            uniform_cond_samples[var] = samples
            conditioning_vars = set(conditioning_vars)

        return rng, conditioning_vars, conditioning_clean, uniform_cond_samples

    def _sampling_order(
        self, conditioning_vars: set[int] | None = None
    ) -> list[tuple[RootNode | ChildNode, bool]]:
        """
        Determine the optimal sampling order for the vine copula.

        The sampling order is derived using a top-down elimination strategy that is
        analogous to finding a perfect elimination ordering for chordal graphs. In our
        vine copula, each conditional density is expressed as a product of bivariate
        copula densities and univariate margins. The algorithm starts with the deepest
        tree (i.e., the tree with the largest conditioning sets) and selects a variable
        from its conditioned set. This variable is then marginalized out, effectively
        generating a new sub-vine with a new deepest node. This process is repeated
        until the first tree level is reached and all n variables have been ordered.

        Choosing the optimal sampling order in this manner simplifies the inversion of
        conditional CDFs, thereby improving numerical stability. At each elimination
        step, among the candidate variables, the one that appears least frequently in
        the conditioning sets of shallower trees is chosen, ensuring that variables
        critical for conditional sampling occupy central roles.

        Parameters
        ----------
        conditioning_vars : set[int], optional
            A set of asset indices for which conditioning samples are provided.
            If specified, these assets will be prioritized during the sampling order
            determination.

        Returns
        -------
        sampling_order : list[tuple(Node, bool | None)]
            A list of tuples representing the optimal sampling order. Each tuple
            contains:
              - A Node object corresponding to an asset or an edge in the vine.
              - A boolean flag indicating whether the left branch is used for sampling
               at that node, or None if the node is the root.
        """
        conditioning_vars = conditioning_vars or set()
        sampling_order: list[tuple[RootNode | ChildNode, bool | None]] = []
        n_assets = self.n_features_in_
        conditioning_counts = self._conditioning_count()

        edges = self.trees_[-1].edges
        if len(edges) == 1:
            edge = edges[0]
            is_left = _is_left_branch(
                edge=edge,
                conditioning_vars=conditioning_vars,
                conditioning_counts=conditioning_counts[-1],
            )
            sampling_order.append((edge.ref_node, is_left))
        else:
            # For truncated trees, select edges minimizing conditioning counts.
            remaining = edges
            visited: set[Edge] = set()
            prev_visited: set[Edge] = set()
            edge, is_left = None, None
            while remaining:
                selected = []
                costs = []
                for edge in remaining:
                    c1 = len(edge.node1.edges - prev_visited) == 1
                    c2 = len(edge.node2.edges - prev_visited) == 1
                    if c1 or c2:
                        if c1 and c2:
                            # Last node
                            is_left = _is_left_branch(
                                edge=edge,
                                conditioning_vars=conditioning_vars,
                                conditioning_counts=conditioning_counts[-1],
                            )
                        else:
                            is_left = c1
                        costs.append(
                            conditioning_counts[
                                -1, edge.cond_sets.conditioned[0 if is_left else 1]
                            ]
                        )
                        selected.append((edge.ref_node, is_left))
                        visited.add(edge)
                remaining = [x for x in remaining if x not in visited]
                # Sort selected nodes by cost and add to sampling order.
                ordered_selected = sorted(
                    zip(costs, selected, strict=True), key=lambda pair: pair[0]
                )
                sampling_order.extend(node for _, node in ordered_selected)
                prev_visited = visited.copy()

        # Marginalization: Peeling off the tree level by level.
        # We have two valid choices (two conditioned vars: left or right),
        # we chose the conditioned var with the less conditioning cost
        i = 2
        while True:
            node = edge.node2 if is_left else edge.node1
            if isinstance(node, RootNode):
                sampling_order.append((node, None))
                break
            else:
                edge = node.ref
                is_left = _is_left_branch(
                    edge=edge,
                    conditioning_vars=conditioning_vars,
                    conditioning_counts=conditioning_counts[-i],
                )
                sampling_order.append((node, is_left))
            i += 1

        # Replace nodes with conditioning samples where applicable.
        if conditioning_vars:
            for i, (node, is_left) in enumerate(sampling_order):
                node_var = node.get_var(is_left) if is_left is not None else node.ref
                if node_var in conditioning_vars:
                    sampling_order[i] = (self.trees_[0].nodes[node_var], None)

        sampling_order = sampling_order[::-1]
        if not (len(set(sampling_order)) == len(sampling_order) == n_assets):
            raise ValueError(
                "Sampling order computation failed: ordering is not unique or complete."
            )
        return sampling_order

    @property
    def fitted_repr(self) -> str:
        """String representation of the fitted Vine Copula."""
        skv.check_is_fitted(self)
        lines = []
        if self.fit_marginals:
            lines.append("Root Nodes")
            lines.append("-" * 10)
            for node, dist in zip(
                self.trees_[0].nodes, self.marginal_distributions_, strict=True
            ):
                lines.append(f"{node}: {dist.fitted_repr}")
            lines.append("")

        for tree in self.trees_:
            lines.append(str(tree))
            lines.append("-" * len(str(tree)))
            for edge in tree.edges:
                lines.append(str(edge))
            lines.append("")

        result_string = "\n".join(lines)
        return result_string

    def display_vine(self):
        """Display the vine trees and fitted copulas.
        Prints the structure of each tree and the details of each edge.
        """
        print(self.fitted_repr)

    def plot_marginal_distributions(
        self,
        X: npt.ArrayLike | None = None,
        conditioning: dict[int | str : float | tuple[float, float] | npt.ArrayLike]
        | None = None,
        subset: list[int | str] | None = None,
        n_samples: int = 500,
        percentile_cutoff: float | None = None,
        title: str = "Vine Copula Marginal Distributions",
    ) -> go.Figure:
        """
        Plot overlaid marginal distributions.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_assets), optional
            Historical data where each column corresponds to an asset.

        conditioning : dict[int | str, float | tuple[float, float] | array-like], optional
            A dictionary specifying conditioning information for one or more assets.
            The dictionary keys are asset indices or names, and the values define how
            the samples are conditioned for that asset. Three types of conditioning
            values are supported:

            1. **Fixed value (float):**
               If a float is provided, all samples are generated under the condition
               that the asset takes exactly that value.

            2. **Bounds (tuple of two floats):**
               If a tuple `(min_value, max_value)` is provided, samples are generated
               under the condition that the asset's value falls within the specified
               bounds. Use `-np.Inf` for no lower bound or `np.Inf` for no upper bound.

            3. **Array-like (1D array):**
               If an array-like of length `n_samples` is provided, each sample is
               conditioned on the corresponding value in the array for that asset.

            When using conditional sampling, it is recommended that the
            assets you condition on are set as central during the vine copula
            construction. This can be specified via the `central_assets` parameter in
            the vine copula instantiation.

        subset : list[int | str], optional
            Indices or names of assets to include in the plot. If None, all assets are
            used.

        n_samples : int, default=500
            Number of samples used to control the density and readability of the plot.
            If `X` is provided and contains more than `n_samples` rows, a random
            subsample of size `n_samples` is selected. Conversely, if `X` has fewer
            rows than `n_samples`, the value is adjusted to match the number of rows in
            `X` to ensure balanced visualization.

        percentile_cutoff : float, default=None
            Percentile cutoff for tail truncation (percentile), in percent.
            If a float p is provided, the distribution support is truncated at
            the p-th and (100 - p)-th percentiles.
            If None, no truncation is applied (uses full min/max of returns).

        title : str, default="Vine Copula Marginal Distributions"
            The title for the plot.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            A figure with overlaid univariate distributions for each asset.
        """
        n_assets = self.n_features_in_
        subset = subset or list(range(n_assets))
        if X is not None:
            X = np.asarray(X)
            if X.ndim != 2:
                raise ValueError("X should be an 2D array")
            if X.shape[1] != n_assets:
                raise ValueError(f"X should have {n_assets} columns")
            if X.shape[0] > n_samples:
                # We subsample for improved graph readability
                rng = sku.check_random_state(self.random_state)
                indices = rng.choice(
                    np.arange(X.shape[0]), size=n_samples, replace=False
                )
                X = X[indices, :]
            else:
                # We want same proportion as X to have a balanced graph
                n_samples = X.shape[0]

        samples = self.sample(n_samples=n_samples, conditioning=conditioning)
        colors = px.colors.qualitative.Plotly

        traces: list[go.Scatter] = []
        for i, s in enumerate(subset):
            visible = True if i == 0 else "legendonly"
            color = colors[i % len(colors)]
            asset = self.feature_names_in_[s]

            traces.append(
                kde_trace(
                    x=samples[:, s],
                    sample_weight=None,
                    percentile_cutoff=percentile_cutoff,
                    name=f"{asset} Generated",
                    line_color=color,
                    fill_opacity=0.17,
                    line_dash="solid",
                    line_width=1,
                    visible=visible,
                )
            )

            if X is not None:
                traces.append(
                    kde_trace(
                        x=X[:, s],
                        sample_weight=None,
                        percentile_cutoff=percentile_cutoff,
                        name=f"{asset} Empirical",
                        line_color=color,
                        fill_opacity=0.17,
                        line_dash="dash",
                        line_width=1.5,
                        visible=visible,
                    )
                )

        fig = go.Figure(data=traces)
        fig.update_layout(
            title=title,
            xaxis_title="x",
            yaxis_title="Probability Density",
        )
        fig.update_xaxes(
            tickformat=".0%",
        )
        return fig

    @contextlib.contextmanager
    def _count_node_visits(self):
        """A context manager to enable counting node visits within the tree.
        Temporarily enables node visit counting for the duration of the context.
        After the block is executed, the original state is restored.
        """
        for tree in self.trees_:
            tree.is_count_visits = True
        try:
            yield
        finally:
            for tree in self.trees_:
                tree.is_count_visits = False
            self.clear_cache(clear_count=False)


def _is_left_branch(
    edge: Edge, conditioning_vars: set[int], conditioning_counts: np.ndarray
) -> bool:
    """
    Determine whether the left branch should be followed during the elimination ordering
    (tree peeling).

    Parameters
    ----------
    edge : Edge
        The edge for which to decide the branch.

    conditioning_vars : set[int]
        Set of asset indices with conditioning samples.

    conditioning_counts : np.ndarray
        Array of conditioning counts for each asset.

    Returns
    -------
    bool
        True if the left branch is preferred, False otherwise.
    """
    v1, v2 = edge.cond_sets.conditioned
    if v1 in conditioning_vars and v2 not in conditioning_vars:
        return False
    if v1 not in conditioning_vars and v2 in conditioning_vars:
        return True
    return conditioning_counts[v1] <= conditioning_counts[v2]


def _propagate_samples(X_rand, sampling_order, conditioning_vars, uniform_cond_samples):
    """Propagate samples through the vine structure bottom-up following the
    elimination strategy (tree peeling) given by the Node orders and whether the next
    Node will on the right or left branch.

    If `is_count_visits` is activated, we only record the number of Node visits in
    each Node. This count is used for optimally clearing cache during sampling.
    """
    is_count = sampling_order[0][0].tree.is_count_visits

    if not is_count:
        X_rand = iter(X_rand)

    # Initialize samples for each node according to the sampling order.
    queue: deque[tuple[RootNode | ChildNode, bool]] = deque()
    for node, is_left in sampling_order:
        node_var = node.get_var(is_left) if is_left is not None else node.ref
        if is_count:
            init_samples = np.array([np.nan])
        else:
            if node_var in conditioning_vars:
                init_samples = uniform_cond_samples[node_var]
            else:
                init_samples = next(X_rand)

        # Avoid Inf
        init_samples = np.clip(
            init_samples,
            a_min=_UNIFORM_SAMPLE_EPSILON,
            a_max=1 - _UNIFORM_SAMPLE_EPSILON,
        )

        if isinstance(node, RootNode):
            node.pseudo_values = init_samples
        else:
            queue.append((node, is_left))
            if is_left:
                node.u = init_samples
            else:
                node.v = init_samples

    while queue:
        node, is_left = queue.popleft()
        edge = node.ref
        if isinstance(edge.node1, RootNode):
            if is_left:
                x = np.stack([node.u, edge.node2.pseudo_values]).T
                edge.node1.pseudo_values = _inverse_partial_derivative(
                    edge, x, is_count
                )
            else:
                x = np.stack([node.v, edge.node1.pseudo_values]).T
                edge.node2.pseudo_values = _inverse_partial_derivative(
                    edge, x, is_count
                )
        else:
            is_left1, is_left2 = edge.node1.ref.shared_node_is_left(edge.node2.ref)
            if is_left:
                x = np.stack([node.u, edge.node2.v if is_left2 else edge.node2.u]).T
                u = _inverse_partial_derivative(edge, x, is_count)
                if is_left1:
                    edge.node1.v = u
                else:
                    edge.node1.u = u
                queue.appendleft((edge.node1, not is_left1))
            else:
                x = np.stack([node.v, edge.node1.v if is_left1 else edge.node1.u]).T
                u = _inverse_partial_derivative(edge, x, is_count)
                if is_left2:
                    edge.node2.v = u
                else:
                    edge.node2.u = u
                queue.appendleft((edge.node2, not is_left2))


def _inverse_partial_derivative(
    edge: Edge, X: np.ndarray, is_count: bool
) -> np.ndarray:
    """Inverse partial derivative of an Edge copula."""
    if is_count:
        return np.array([np.nan])
    return edge.copula.inverse_partial_derivative(X)
